import json
import subprocess
import os
import shlex
import sys
import signal
import time
import optparse
from pipes import quote


import gettext
from gettext import gettext as _
gettext.textdomain('fsgamer')

from fsgamer_lib import get_version, get_data_path


class DependencyError(Exception): pass

BINPATH = os.environ.get("FSGAMER_BIN", sys.argv[0])

# Quick hack for during development using quickly
QUICKLY_RUN_PATH = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(
                                os.path.realpath(__file__)))), "quicklyrun.sh")
if os.path.exists(QUICKLY_RUN_PATH):
    BINPATH = QUICKLY_RUN_PATH


HOMEDIR = os.path.expanduser("~")

TITLE_TRANS = ''.join(chr(c) if chr(c).isupper() or chr(c).islower() else '_' for c in range(256))

def _alphanum(s):
    return s.strip().translate(TITLE_TRANS).replace("_", "")

class shellbuilder:
    XINIT_SCRIPT = "DISPLAY=:%(display)i.0\n%(xinit)s %(cmd)s  -- :%(display)i"
    @staticmethod
    def xinit(cmd):

        xinit_path = shell_run.which("xinit")
        cklaunch_path = shell_run.which("ck-launch-session")

        xinit = "%s %s" % (xinit_path, cklaunch_path)

        # Wrap script
        script = shellbuilder.XINIT_SCRIPT % {
            'xinit': xinit,
            'display': 1,
            'cmd': cmd.strip(),
        }

        return script

    FSGWM_SCRIPT = "%(base)s --wm --exec=%(exec)s --opts=%(opts)s"
    @staticmethod
    def fsgwm(cmd, opts, openbox=True):
        """
        Generates path to run windows manager, auto-execing given command path
        """
        import openbox
        d = {
            'exec': cmd,
            'opts': opts,
        }

        if openbox and openbox.ENABLED:
            d['openbox'] = None
        else:
            d['wm'] = None

        return unparse_options(d)


TMP_DIR = "/tmp/fsgamer/"
def tmp_file_path(path):
    return os.path.join(TMP_DIR, path)

def write_tmp_file(path, data):
    path = os.path.join(TMP_DIR, path)
    dirpath = os.path.dirname(path)
    try:
        os.makedirs(dirpath)
    except OSError:
        pass

    open(path, "w").write(data)
    return path


class ShellRun:
    TMP_DIR = "/tmp/fsgamer/shell_run/"
    def __init__(self):
        self.processes = []
        self.defunct = []

    def killall(self):
        pass
        #for process in self.processes:
        #    process.terminate()
        #    time.sleep(2)
        #    process.kill()
        #for process in self.processes:
        #    process.wait()
        #self.defunct.extend(self.processes)
        #self.processes = []
        #os.kill(process.pid, signal.SIGKILL)
        #os.killpg(0, signal.SIGKILL) # kill all processes in my group
        #self.process.wait()

    def replace_killall_fsgamer(self):
        #if os.fork() == 0:
        killall = self.which('killall')
        os.execv(killall, [killall, "fsgamer ck-launch-session"])

    def replace_kill_everything(self, exe=None):
        hitlist = [
                    "ck-launch-session",
                    "openbox",
                    "fsgamer_",
                    BINPATH,
                ]

        if exe:
            hitlist.append(exe)

        processes = self.get_processes()

        pid_hitlist = []
        for pid, cmdline in processes.iteritems():
            for item in hitlist:
                if item in cmdline:
                    pid_hitlist.append(pid)

        script = """
            killall fsgamer
            kill IDS
            sleep 5
            kill -9 IDS
        """.replace("IDS", " ".join(map(str, pid_hitlist)))

        if not pid_hitlist:
            return
        self(script, replace=True)

    def sh_from_script(self, script):
        script_text = "#!/bin/bash\n%s" % script
        fn = "fsgamer_%i" % len(script)
        path = os.path.join(self.TMP_DIR, fn)
        try:
            os.makedirs(self.TMP_DIR)
        except OSError:
            pass

        open(path, "w").write(script_text)

        os.chmod(path, 0777)
        assert os.path.exists(path)

        return path

    def __call__(self, script, replace=False):
        path = self.sh_from_script(script)
        if not replace:
            process = subprocess.Popen(path,
                                    shell=True,
                                    bufsize=-1)
            self.processes.append(process)
            return process
        else:
            os.execv(path, [path])
        #script_args = shlex.split(script)
        #return subprocess.Popen(script_args, shell=True, bufsize=-1)

    def which(self, cmd):
        path = subprocess.check_output("which %s"%cmd, shell=True)
        if not path:
            raise DependencyError("%s was not found" % cmd)
        return path.strip()

    def get_processes(self):
        pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
        r = {}
        for pid in pids:
            path = os.path.join('/proc', pid, 'cmdline')
            try:
                r[pid] = open(path, 'rb').read()
            except OSError: pass
            except IOError: pass
        return r

    def is_running(self, search, exact=False):
        processes = self.get_processes()
        for pid, cmd in processes.iteritems():
            if exact:
                if _alphanum(search) == _alphanum(cmd):
                    return pid
            elif search in cmd:
                return pid
        return False

    def return_when_stopped_running(self, exe):
        found_pid = None

        while not found_pid:
            time.sleep(10)
            processes = self.get_processes()
            found_pid = self.is_running(exe, exact=True)

        pid_path = os.path.join('/proc', found_pid)

        while True:
            if not os.path.exists(pid_path):
                # process ended!
                return
            time.sleep(3)

    def xinit_fsgwm(self, cmd, opts):
        """
        This is the main function where all the magic is:

        It runs (blocking) a command using the FSGamer windows manager in a new
        xinit buffer wrapped around ck-launch-session, passing to the windows
        manager (via JSON) the dictionary of opts
        """
        script = []

        if opts.get('notifications_tts'):
            # If we have TTS notifications enabled, then add that to the script
            script.append(unparse_options({'tts': None}) + " &")

        if not opts.get('keep_going'):
            # Run kill-pact process to kill everything if the child process
            # fails to start
            d = {'killpact': None, 'exec': cmd}
            script.append(unparse_options(d) + " &")

        script.append(
                shellbuilder.xinit(shellbuilder.fsgwm(cmd, opts))
            )

        return self("\n".join(script))
        #return subprocess.Popen(script_args, shell=True, bufsize=-1)

shell_run = ShellRun()

def unparse_options(dct, command=None, use_equals=True):
    args = [command or BINPATH]
    for item, val in dct.items():
        if val is None:
            args.append('--%s'%item)
            continue

        if isinstance(val, dict):
            val = json.dumps(val)

        if use_equals:
            args.append("--%s=%s" % (item, quote(str(val))))
        else:
            args.append("--%s %s" % (item, quote(str(val))))

    return " ".join(args)

def parse_options(string=None):
    """Support for command line options"""
    parser = optparse.OptionParser(version="%%prog %s" % get_version())

    parser.add_option(
        "-v", "--verbose", action="count", dest="verbose",
        help=_("Show debug messages (-vv debugs fsgamer_lib also)"))

    parser.add_option(
        "-e", "--exec",   dest="exe", metavar="CMD",
        help=_("Executes given shell command in a new xsession"))

    parser.add_option(
        "-d", "--desktop",  dest="desktop", metavar="FILE.desktop",
        help=_("Path to desktop file responsible to enable GUI"))

    parser.add_option(
        "-o", "--opts",   dest="opts", metavar="JSON",
        help=_("JSON opts format for more controller options"))

    parser.add_option(
        "-w", "--wm", action="store_true", dest="wm",
        help=_("Launch custom windows manager, in current X session, auto-running exec"))

    parser.add_option(
        "-b", "--openbox", action="store_true", dest="openbox",
        help=_("Launch openbox, in current X session, auto-running exec"))

    parser.add_option(
        "-t", "--tts", action="store_true", dest="tts",
        help=_("Speaks notifications via TTS"))

    parser.add_option(
        "-K", "--killall", action="store_true", dest="killall",
        help=_("Kills all processes named 'fsgamer', 'openbox' and exec"))

    parser.add_option(
        "-p", "--killpact", action="store_true", dest="killpact",
        help=_("Killall when exec stops running"))

    if string:
        (options, args) = parser.parse_args(shlex.split(string))
    else:
        (options, args) = parser.parse_args(sys.argv)

    if options.opts:
        # Should be a JSON object at this point
        try:
            print "options.opts", options.opts
            options.opts = json.loads(options.opts)
        except:
            print "Uh oh couldn't load options..."
            options.opts = {}

    return options, args
